import { unified } from 'unified';
import { unifiedLatexFromString, unifiedLatexAstComplier, } from '@unified-latex/unified-latex-util-parse';
import { unifiedLatexAttachMacroArguments, gobbleArguments, } from '@unified-latex/unified-latex-util-arguments';
import { typstMacros, typstStrings } from './macros.js';
export function parseLatex(value) {
    const file = unified()
        .use(unifiedLatexFromString, { mode: 'math' })
        .use(unifiedLatexAstComplier)
        .use(unifiedLatexAttachMacroArguments, {
        macros: {
            vec: { signature: 'm' },
            boldsymbol: { signature: 'm' },
            left: { signature: 'm' },
            right: { signature: 'm' },
        },
    })
        .processSync({ value });
    const content = file.result.content;
    const tree = { type: 'math', content };
    return tree;
}
export * from './macros.js';
export function walkLatex(node) {
    delete node.position;
    if (Array.isArray(node.content)) {
        const content = node.content.map((n) => walkLatex(n));
        let skip = 0;
        const parsed = content.reduce((list, next, i, array) => {
            var _a;
            if (skip > 0) {
                skip -= 1;
                return list;
            }
            if (next.type === 'string' && (next.content === '_' || next.content === '^')) {
                const { args, nodesRemoved } = gobbleArguments(array.slice(i + 1), 'm');
                next.type = 'macro';
                next.args = args;
                skip += nodesRemoved;
            }
            if (next.type === 'macro' &&
                (next.content === 'overbrace' || next.content === 'underbrace')) {
                const { args, nodesRemoved } = gobbleArguments(array.slice(i + 1), 'm');
                if (args[0].content.length === 1 &&
                    args[0].content[0].type === 'macro' &&
                    ((args[0].content[0].content === '^' && next.content === 'overbrace') ||
                        (args[0].content[0].content === '_' && next.content === 'underbrace'))) {
                    next.args = [...((_a = next.args) !== null && _a !== void 0 ? _a : []), ...args[0].content[0].args];
                    skip += nodesRemoved;
                }
            }
            list.push(next);
            return list;
        }, []);
        node.content = parsed;
        return { ...node, content: parsed };
    }
    if (Array.isArray(node.args)) {
        const args = node.args.map((n) => walkLatex(n));
        node.args = args;
        return { ...node, args };
    }
    return node;
}
class State {
    constructor() {
        this._scriptsSimplified = false;
        this._closeToken = [];
        this._value = '';
    }
    get value() {
        return this._value;
    }
    addWhitespace() {
        const lastChar = this.value.slice(-1);
        if (!this._value || lastChar.match(/^(["\s_^{(-])$/))
            return;
        this._value += ' ';
    }
    write(str) {
        if (!str)
            return;
        // This is a bit verbose, but the statements are much easier to read
        if (this._scriptsSimplified && str === '(') {
            this.addWhitespace();
        }
        else if (str.match(/^([}()_^,!])$/)) {
            // Ignore!
        }
        else {
            this.addWhitespace();
        }
        this._scriptsSimplified = false;
        this._value += str;
    }
    openFunction(command) {
        if (command === 'text') {
            this.addWhitespace();
        }
        else {
            this.write(command);
        }
        this._simplify = command === '_' || command === '^';
        this._lastFunction = this._value.length;
        this._value += command === 'text' ? '"' : '(';
        this._closeToken.push(command === 'text' ? '"' : ')');
    }
    closeFunction() {
        this._value += this._closeToken.pop() || ')';
        if (!this._simplify)
            return;
        // We will attempt to change `x_(i)` into `x_i`
        const simple = this._value.slice(this._lastFunction);
        if (simple.length === 3 || simple.match(/^\([a-zA-Z]*\)$/)) {
            this._value = this._value.slice(0, this._lastFunction) + simple.slice(1, -1);
            this._scriptsSimplified = true;
        }
    }
}
function convert(node) {
    if (node.type === 'macro' && typeof node.content === 'string') {
        const result = typstMacros[node.content];
        const converted = typeof result === 'function' ? result(node) : result;
        return converted !== null && converted !== void 0 ? converted : node.content;
    }
    return '';
}
function convertText(text) {
    const result = typstStrings[text];
    return result || text;
}
export function writeTypst(node, state = new State()) {
    if (node.type === 'whitespace') {
        // We are controlling whitespace in the renderer
        return state;
    }
    else if (node.type === 'string') {
        state.write(convertText(node.content));
    }
    else if (Array.isArray(node.content)) {
        node.content.forEach((n) => {
            writeTypst(n, state);
        });
    }
    else if (node.type === 'macro' && Array.isArray(node.args)) {
        const converted = convert(node);
        if (node.args.length === 0) {
            state.write(converted);
            return state;
        }
        state.openFunction(converted);
        node.args
            .filter((n) => {
            if (Array.isArray(n.content) && n.content.length === 0)
                return false;
            return true;
        })
            .forEach((n, i) => {
            if (i !== 0)
                state.write(',');
            writeTypst(n, state);
        });
        state.closeFunction();
    }
    else if (node.type === 'macro' && typeof node.content === 'string') {
        const converted = convert(node);
        state.write(converted !== null && converted !== void 0 ? converted : node.content);
    }
    return state;
}
export function texToTypst(value) {
    const tree = parseLatex(value);
    walkLatex(tree);
    const state = writeTypst(tree);
    return state.value;
}
